@@ -35,6 +35,8 @@ module Network.Wai.Handler.WarpTLS (
     -- * Runner
     , runTLS
     , runTLSSocket
+    , runHTTP2TLS
+    , runHTTP2TLSSocket
     -- * Exception
     , WarpTLSException (..)
     ) where
@@ -42,6 +44,7 @@ module Network.Wai.Handler.WarpTLS (
 #if __GLASGOW_HASKELL__ < 709
 import Control.Applicative ((<$>))
 #endif
+import Control.Applicative ((<|>))
 import Control.Exception (Exception, throwIO, bracket, finally, handle, fromException, try, IOException, onException, SomeException(..))
 import qualified Control.Exception as E
 import Control.Monad (void)
@@ -57,6 +60,7 @@ import Network.Socket.ByteString (sendAll)
 import qualified Network.TLS as TLS
 import qualified Network.TLS.Extra as TLSExtra
 import Network.Wai (Application)
+import Network.Wai.HTTP2 (HTTP2Application)
 import Network.Wai.Handler.Warp
 import Network.Wai.Handler.Warp.Internal
 import System.IO.Error (isEOFError)
@@ -217,21 +221,39 @@ tlsSettingsChainMemory cert chainCerts key = defaultTlsSettings
 
 ----------------------------------------------------------------
 
+-- Composition over two arguments at once; used for runHTTP2\*.
+(.:) :: (c -> d) -> (a -> b -> c) -> a -> b -> d
+f .: g = curry $ f . uncurry g
 
 -- | Running 'Application' with 'TLSSettings' and 'Settings'.
 runTLS :: TLSSettings -> Settings -> Application -> IO ()
-runTLS tset set app = withSocketsDo $
+runTLS tset set = runServeTLS tset set . serveDefault
+
+-- | Run an HTTP\/2-aware server with 'TLSSettings' and 'Settings'.
+runHTTP2TLS :: TLSSettings -> Settings -> HTTP2Application -> Application -> IO ()
+runHTTP2TLS tset set = runServeTLS tset set .: serveHTTP2
+
+runServeTLS :: TLSSettings -> Settings -> ServeConnection -> IO ()
+runServeTLS tset set serve = withSocketsDo $
     bracket
         (bindPortTCP (getPort set) (getHost set))
         sClose
-        (\sock -> runTLSSocket tset set sock app)
+        (\sock -> runServeTLSSocket tset set sock serve)
 
 ----------------------------------------------------------------
 
 -- | Running 'Application' with 'TLSSettings' and 'Settings' using
 --   specified 'Socket'.
 runTLSSocket :: TLSSettings -> Settings -> Socket -> Application -> IO ()
-runTLSSocket tlsset@TLSSettings{..} set sock app = do
+runTLSSocket tset set sock = runServeTLSSocket tset set sock . serveDefault
+
+-- | Run an HTTP\/2-aware server with 'TLSSettings' and 'Settings' using
+--   specified 'Socket'.
+runHTTP2TLSSocket :: TLSSettings -> Settings -> Socket -> HTTP2Application -> Application -> IO ()
+runHTTP2TLSSocket tset set sock = runServeTLSSocket tset set sock .: serveHTTP2
+
+runServeTLSSocket :: TLSSettings -> Settings -> Socket -> ServeConnection -> IO ()
+runServeTLSSocket tlsset@TLSSettings{..} set sock serve = do
     credential <- case (certMemory, keyMemory) of
         (Nothing, Nothing) ->
             either error id <$>
@@ -241,14 +263,14 @@ runTLSSocket tlsset@TLSSettings{..} set sock app = do
             key <- maybe (S.readFile keyFile) return mkey
             either error return $
               TLS.credentialLoadX509ChainFromMemory cert chainCertsMemory key
-    runTLSSocket' tlsset set credential sock app
+    runServeTLSSocket' tlsset set credential sock serve
 
-runTLSSocket' :: TLSSettings -> Settings -> TLS.Credential -> Socket -> Application -> IO ()
-runTLSSocket' tlsset@TLSSettings{..} set credential sock app =
-    runSettingsConnectionMakerSecure set get app
+runServeTLSSocket' :: TLSSettings -> Settings -> TLS.Credential -> Socket -> ServeConnection -> IO ()
+runServeTLSSocket' tlsset@TLSSettings{..} set credential sock serve =
+    runServeSettingsConnectionMakerSecure set get serve
   where
     get = getter tlsset sock params
-    params = TLS.ServerParams {
+    params = def { -- TLS.ServerParams
         TLS.serverWantClientCert = tlsWantClientCert
       , TLS.serverCACertificates = []
       , TLS.serverDHEParams      = Nothing
@@ -258,12 +280,13 @@ runTLSSocket' tlsset@TLSSettings{..} set credential sock app =
       }
     -- Adding alpn to user's tlsServerHooks.
     hooks = tlsServerHooks {
-        TLS.onALPNClientSuggest = Just alpn
+        TLS.onALPNClientSuggest = TLS.onALPNClientSuggest tlsServerHooks <|>
+          (if settingsHTTP2Enabled set then Just alpn else Nothing)
       }
     shared = def {
         TLS.sharedCredentials = TLS.Credentials [credential]
       }
-    supported = TLS.Supported {
+    supported = def { -- TLS.Supported
         TLS.supportedVersions       = tlsAllowedVersions
       , TLS.supportedCiphers        = tlsCiphers
       , TLS.supportedCompressions   = [TLS.nullCompression]
@@ -439,7 +462,12 @@ plainHTTP TLSSettings{..} s bs0 = case onInsecure of
                 }
         return (conn'', TCP)
     DenyInsecure lbs -> do
-        sendAll s "HTTP/1.1 200 OK\r\nContent-Type: text/plain\r\n\r\n"
+        -- FIXME: what about HTTP/2?
+        -- http://tools.ietf.org/html/rfc2817#section-4.2
+        sendAll s "HTTP/1.1 426 Upgrade Required\
+        \r\nUpgrade: TLS/1.0, HTTP/1.1\
+        \r\nConnection: Upgrade\
+        \r\nContent-Type: text/plain\r\n\r\n"
         mapM_ (sendAll s) $ L.toChunks lbs
         sClose s
         throwIO InsecureConnectionDenied
